Attribute VB_Name = "CampanusPositions"
' This module computes Topocentric Planetary Positions and Campanus Houses

' --- Types and Data, private usage, but VBA doesn't allow them to be placed after public interface
Private Type InputParameters
  jd_ut As Double
  lon As Double
  lat As Double
  alt As Double
End Type

Private Type PlanetCoord
  lon   As Double
  lat   As Double
  ra    As Double
  decl  As Double
  mp    As Double
  oa    As Double
  pol   As Double
  lon_proj As Double
End Type

Private planets(0 To 9) As PlanetCoord
Private p As PlanetCoord
Private i As InputParameters
Private ramc As Double
Private houses(0 To 12) As Double
Private errMsg As String

Private Const Deg2Arc As Double = 1.74532925199433E-02
Private Const Campanus As Long = 67  ' = ASCII-Code für "C"


' --- Public Interface
Public Function getLon(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getLon = getData(ipl, "lon", jd_ut, lon, lat, alt)
End Function

Public Function getLat(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getLat = getData(ipl, "lat", jd_ut, lon, lat, alt)
End Function

Public Function getRA(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getRA = getData(ipl, "ra", jd_ut, lon, lat, alt)
End Function

Public Function getDecl(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getDecl = getData(ipl, "decl", jd_ut, lon, lat, alt)
End Function

Public Function getMP(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getMP = getData(ipl, "mp", jd_ut, lon, lat, alt)
End Function

Public Function getLonProj(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getLonProj = getData(ipl, "lon_proj", jd_ut, lon, lat, alt)
End Function

Public Function getPol(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getPol = getData(ipl, "pol", jd_ut, lon, lat, alt)
End Function

Public Function getOA(ipl As Long, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  getOA = getData(ipl, "oa", jd_ut, lon, lat, alt)
End Function

Public Function getHouse(cusp As Long, jd_ut As Double, lon As Double, lat As Double, Optional alt As Double = -100000) As Double
  If alt < -10000 Then
    getHouse = getData(cusp, "house", jd_ut, lon, lat, i.alt)
  Else
    getHouse = getData(cusp, "house", jd_ut, lon, lat, alt)
  End If
End Function

Public Function getRAMC(jd_ut As Double, lon As Double) As Double
  getRAMC = swe_sidtime(jd_ut) * 15 + lon
End Function

Public Function getError() As String
  getError = errMsg
End Function



' --- Implementation

Sub Test()
  Dim l As Double
  l = getLon(1, 2452232, 7, 52, 0)
  MsgBox l
End Sub

Private Sub ComputePlanets(jd_ut As Double, lon As Double, lat As Double, altitude As Double)
  Dim ipl As Long, ee As Double, xx(6) As Double, serr As String
  
' Ekliptikschiefe des Datums
  iflag = 0
  swe_calc_ut jd_ut, SE_ECL_NUT, iflag, xx(0), serr
  setErrorFrom serr
  
  ee = xx(0) * Deg2Arc
  
' Ortskoordinaten der Swiss Ephemeris bekanntmachen
  swe_set_topo lon, lat, altitude
  
' Sternzeit
  ramc = getRAMC(jd_ut, lon)
  
  For ipl = 0 To 9
    ComputeSinglePlanet ipl, jd_ut, lon, lat, altitude, ramc, ee, planets(ipl)
  Next

End Sub


Private Function trimString(ByVal inString As String) As String
  Dim length As Long
  length = InStr(inString, Chr$(0)) - 1
  If length >= 0 Then
    trimString = Left$(inString, length)
  Else
    trimString = ""
  End If
End Function

Sub setErrorFrom(serr As String)
  Dim s
  s = trimString(serr)
  If s <> "" Then
    errMsg = s
  End If
End Sub

Private Sub ComputeSinglePlanet(ipl As Long, jd_ut As Double, lon As Double, lat As Double, _
                                altitude As Double, ramc As Double, ee As Double, ByRef result As PlanetCoord)
  
  Dim xx(6) As Double, serr As String, iflag As Long, iret As Long
  Dim kpd As Double, x As Double
    
' Planet in (topozentrischen) Ekliptikalkoordinaten berechnen
  iflag = SEFLG_TOPOCTR
  serr = String(255, 0)
  iret = swe_calc_ut(jd_ut, ipl, iflag, xx(0), serr)
  setErrorFrom serr
  
  result.lon = xx(0)
  result.lat = xx(1)
  
' Planet in Äquatorialkoordinaten (RA, Deklination) berechnen
  iflag = SEFLG_EQUATORIAL + SEFLG_TOPOCTR
  swe_calc_ut jd_ut, ipl, iflag, xx(0), serr
  If serr <> "" Then errMsg = serr
  
  result.ra = xx(0)
  result.decl = xx(1)
  
' Indikator für "über/unter dem Horizont", siehe Primärdirektionen, Formel 21
  x = Tan(xx(1) * Deg2Arc) * Tan(lat * Deg2Arc) + Cos((xx(0) - ramc) * Deg2Arc)
  
' Siehe Primärdirektionen, Formel 44
  result.mp = -Atn((Cos(lat * Deg2Arc) * Cos(result.decl * Deg2Arc) * Cos((result.ra - ramc) * Deg2Arc) _
               + Sin(lat * Deg2Arc) * Sin(result.decl * Deg2Arc)) _
            / (Cos(result.decl * Deg2Arc) * Sin((result.ra - ramc) * Deg2Arc)))
  
' Wert in den Bereich zwischen 0 und 360 bringen
  If result.mp < 0 Then result.mp = result.mp + 2 * Application.Pi
  
' Korrektur der Mehrdeutigkeit des ArcusTangens
  If x < 0 And result.mp > Application.Pi Then result.mp = result.mp - Application.Pi
  If x > 0 And result.mp < Application.Pi Then result.mp = result.mp + Application.Pi
  
' Polhöhe gemäss Primärdirektionen, Formel 40
  result.pol = Asn(Cos(result.mp) * Sin(lat * Deg2Arc))
  
' OA und Distanz zu KP (KPD) gemäss Primärdirektionen, Formel 41
  kpd = -Atn(Cos(lat * Deg2Arc) / Tan(result.mp))
  If result.mp < Application.Pi Then kpd = kpd + Application.Pi
  result.oa = ramc * Deg2Arc + kpd
  If result.oa < 0 Then result.oa = result.oa + 2 * Application.Pi
  If result.oa > 2 * Application.Pi Then result.oa = result.oa - 2 * Application.Pi

  result.lon_proj = Atn(Sin(result.oa) / (Cos(ee) * Cos(result.oa) - Sin(ee) * Tan(result.pol))) / Deg2Arc

' Ergebnis liegt "in derselben Hälfte" wie oa
  If result.lon_proj < 0 Then result.lon_proj = result.lon_proj + 360
  If result.lon_proj < 180 And result.oa > Application.Pi Then result.lon_proj = result.lon_proj + 180
  If result.lon_proj > 180 And result.oa < Application.Pi Then result.lon_proj = result.lon_proj - 180

  result.oa = result.oa / Deg2Arc
  result.pol = result.pol / Deg2Arc
  result.mp = result.mp / Deg2Arc

End Sub

Private Function getData(ipl As Long, what As String, jd_ut As Double, lon As Double, lat As Double, alt As Double) As Double
  
  Dim ascmc(0 To 10) As Double
  
  If jd_ut <> i.jd_ut Or _
     lon <> i.lon Or _
     lat <> i.lat Or _
     alt <> i.alt Then
     errMsg = ""
     ComputePlanets jd_ut, lon, lat, alt
    If jd_ut <> i.jd_ut Or _
       lon <> i.lon Or _
       lat <> i.lat Then
       swe_houses jd_ut, lat, lon, Campanus, houses(0), ascmc(0)
    End If
    i.jd_ut = jd_ut
    i.lon = lon
    i.lat = lat
    i.alt = alt
  End If
  
  
  Select Case what
    Case "lon"
      getData = planets(ipl).lon
    Case "lat"
      getData = planets(ipl).lat
    Case "ra"
      getData = planets(ipl).ra
    Case "decl"
      getData = planets(ipl).decl
    Case "lon_proj"
      getData = planets(ipl).lon_proj
    Case "mp"
      getData = planets(ipl).mp
    Case "pol"
      getData = planets(ipl).pol
    Case "oa"
      getData = planets(ipl).oa
    Case "house"
      getData = houses(ipl)
  End Select
  
  
End Function









